
#include <linux/types.h>
#include "ext2_fs_i.h"
#include <linux/types.h>
#include <linux/capability.h>
#include <linux/sched.h>
#include <linux/fs.h>


#define ALLOC_SIZE(i)  i->u.ext2_i.i_alloc_size
#define ALLOC_INUSE(i) i->u.ext2_i.i_alloc_inuse
#define ALLOC_PARENT(i) i->u.ext2_i.i_alloc_parent
#define ALLOC_AVAIL(i) (ALLOC_SIZE(i)-ALLOC_INUSE(i))

#if 0
#define debug(fmt,args...) printk( KERN_WARNING fmt,args)
#else
#define debug(fmt,args...) ;
#endif


int  allocfs_list_alloc( struct inode *i, int *size, int *inuse, int *parent )
{
	*size = ALLOC_SIZE(i);
	*inuse = ALLOC_INUSE(i);
	*parent = ALLOC_PARENT(i);
	return 0;
}

int  allocfs_set_alloc( struct inode *i, int size, int inuse, int parent )
{
	ALLOC_SIZE(i) = size;
	ALLOC_INUSE(i) = inuse;
	ALLOC_PARENT(i) = parent;

	mark_inode_dirty(i);

	return 0;
}

extern int ext2_empty_dir( struct inode *i );

int  allocfs_make_alloc( struct inode *i, int newsize )
{
	struct inode *parent;
	int change;

	debug("inode %lu allocate %d blocks\n",i->i_ino,newsize);

	/* Allocation size must be positive */
	if(newsize<0) return -EINVAL;

	/* This only works on a directory. */
	if(!S_ISDIR(i->i_mode)) return -ENOTDIR;

	if(ALLOC_SIZE(i)==0) {
		/* If not an allocation, must be an empty directory */
		if(!ext2_empty_dir(i)) return -ENOTEMPTY;
		change = newsize;
	} else {
		/* If already an allocation, we may modify it. */
		change = newsize-ALLOC_SIZE(i);
		/* Cannot descend below existing size without deleting. */
		if(newsize<ALLOC_INUSE(i)) return -EDQUOT;
	}

	parent = iget(i->i_sb,ALLOC_PARENT(i));
	if(!parent) {
		debug("couldn't find allocation parent of inode %lu\n",i->i_ino);
		iput(parent);
		return -EIO;
	}

	if(change>0) {
		if(ALLOC_AVAIL(parent)<change) {
			debug("parent only has %d blocks available (need %d)\n",ALLOC_AVAIL(parent),change);
			iput(parent);
			return -EDQUOT;
		}
	}

	ALLOC_INUSE(parent) += change;
	ALLOC_SIZE(i) = newsize;
	if(change==newsize) ALLOC_INUSE(i) = 0;

	mark_inode_dirty(parent);
	mark_inode_dirty(i);

	iput(parent);

	return 0;
}

void allocfs_transfer_state_to_child( struct inode *dir, struct inode *inode )
{
	ALLOC_SIZE(inode) = 0;
	ALLOC_INUSE(inode) = 0;
	if(ALLOC_SIZE(dir)) {
		ALLOC_PARENT(inode) = dir->i_ino;
	} else {
		ALLOC_PARENT(inode) = ALLOC_PARENT(dir);
	}
	mark_inode_dirty(inode);
}

static int allocate_block( struct inode *i )
{
	if(ALLOC_INUSE(i) < ALLOC_SIZE(i) ) {
		ALLOC_INUSE(i)++;
		debug("allocate: inode %lu size %d inuse %d\n",i->i_ino,ALLOC_SIZE(i),ALLOC_INUSE(i));
		mark_inode_dirty(i);
		return 1;
	} else {
		return 0;
	}
}

int allocfs_allocate_block( struct inode *i )
{
	if(ALLOC_SIZE(i)!=0) {
		return allocate_block(i);
	} else if(ALLOC_PARENT(i)!=0) {
		struct inode *parent;
		debug("parent of %lu is %d\n",i->i_ino,ALLOC_PARENT(i));
		parent = iget(i->i_sb,ALLOC_PARENT(i));
		if(parent) {
			int result;
			result = allocate_block(parent);
			iput(parent);
			return result;
		} else {
			return 0;

		}
	} else {
		return 1;
	}
}

static void free_blocks( struct inode *i, int nblocks )
{
	if(ALLOC_INUSE(i)>=nblocks) {
		ALLOC_INUSE(i) -= nblocks;
	} else {
		ALLOC_INUSE(i) = 0;
	}
	mark_inode_dirty(i);
	debug("free: inode %lu size %d inuse %d\n",i->i_ino,ALLOC_SIZE(i),ALLOC_INUSE(i));
}

void allocfs_free_blocks( struct inode *i, int nblocks )
{
	if(ALLOC_SIZE(i)!=0) {
		free_blocks(i,nblocks);
	} else if(ALLOC_PARENT(i)!=0) {
		struct inode *parent = iget(i->i_sb,ALLOC_PARENT(i));
		if(parent) {
			free_blocks(parent,nblocks);
			iput(parent);
		}
	} else {
		// do nothing
	}
}

void allocfs_free_inode( struct inode *i )
{
	if(ALLOC_SIZE(i)==0) {
		debug("inode %lu is not an allocation\n",i->i_ino);
	} else if(ALLOC_PARENT(i)!=0) {
		struct inode *parent = iget(i->i_sb,ALLOC_PARENT(i));
		if(parent) {
			free_blocks(parent,ALLOC_SIZE(i));
			debug( "freed %d blocks in parent %lu\n",ALLOC_SIZE(i),parent->i_ino);
			iput(parent);
		}
	}
}

